#ifndef armv6_m__tl_h
#define armv6_m__tl_h

/*
  armv6-m/tl.h 
  Copyright 2014 Marc Prager
 
  This file is part of the c-lpc8 library.
  c-lpc8 is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 
  c-lpc8 is published in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License along with c-lpc8.
  If not see <http://www.gnu.org/licenses/>
 */

/** @file
 * @brief Thread loops - a simple context switcher between a thread in a loop and a second context.
 *
 * This module tl is intended for use in interrupt service routines to synchronized a sequential control flow and the
 * interrupt. It's based on cooperative multitasking. Focus is on simplicity.
 * tl works with infinite loops only - all threads loops have to continue working indefinitely. Any number of thread
 * loops is possible. Scheduling has to be done explicitely. The intended usage is as follows: main() inititializes
 * the TlContext structures (stack and nonterminating loop function) and calls tlCreate() for ever thread loop. That
 * 'starts' all threads and they run - one after the other - until the first tlYield() in every loop. Upon tlYield()
 * control is transferred back to main(). It is then the responsibility of main() (or an ISR!) to wake up the needed
 * thread loops by invoking tlCall() for the respective TlContext structure.
 * A simple round robin scheduler can be implemented by invoking tlCall() for each thread loop inside the main loop.
 */

/** This data structure is shared between the two contexts an contains all information that is required for context
 * switching.
 */
typedef struct {
	void*	spThread;	///< has to be initialized to a valid (thread) stack pointer.
	void*	spCaller;	///< set/re-set by tlCreate/tlCall.
} TlContext;

// #define NORETURN __attribute__((noreturn,nothrow));
#define NORETURN	// line above doesn't work, unfortunately

/** A thread loop must be of this type. The function MUST NOT return, it has to be a loop.
 */
typedef void TlLoop (TlContext *ctx) NORETURN;

/** Creates a new thread loop. This function is typically called from the main thread. Currently it does not return
 * until the first tlYield() is executed inside the new thread.
 * @param ctx the context switch information storage
 * @param loop the thread function (not returning).
 */
void tlCreate (TlContext *ctx, TlLoop *loop);

/** This function yields control from the thread to the caller.
 * @param ctx the context switch information storage
 */
void tlYield (TlContext *ctx);

/** This function continues the thread at the latest yield() call. It will return on the next tlYield() in the thread.
 * @param ctx the context switch information storage
 */
void tlCall (TlContext *ctx);

#endif

