/*
  Copyright 2014 Marc Prager

  Multithreading example.
 */

#include <armv6-m/tl.h>		// context switcher for Cortex-M0

#include <lpc8/syscon.h>
#include <lpc8/swm.h>
#include <lpc8/usart.h>
#include <lpc8/systickTimer.h>

#include <print.h>
#include <fixedPoint.h>

enum {
	PCLK	= 12*MEGA,
	BAUD	= 115200,
};

void handlerDefault(void) {
	printString("CRASHED!! - STOP.");
	while(true);
}

void sleepMs(TlContext *ctx, int ms) {
	int timeoutMs = systickTimerMs + ms;
	while (timeoutMs-systickTimerMs > 0) tlYield(ctx);
}

__attribute__((naked)) Uint32 getSp(void) {
	asm("mov r0,sp");	// AAPCS: r0 = function return
	asm("bx lr");
}

#define TOP_OF_STACK(s) (&s[sizeof s/sizeof s[0]])

// prototypes: signature enforcement
TlLoop
	threadA,	// printing 'A' every 500ms
	threadB,	// printing 'B<num>' every 2s, num counting up
	threadT,	// complex thread, calling 2 child threads every second
	threadT1,	// printing 'T1', then yield
	threadT2;	// printing 'T2', then yield

void threadA(TlContext *ctx) {
	tlYield(ctx);
	while(true) {
		printChar('A');
		sleepMs(ctx,500);
	}
}

void threadB(TlContext *ctx) {
	tlYield(ctx);
	for (int i=0; ; i++) {
		printChar('B'); printUDec32(i); printChar(' ');
		sleepMs(ctx,2000);
	}
}


void threadT(TlContext *ctx) {
	Uint32	stackT1[64], stackT2[64];	// complex task, needs more threads
	// these stacks reside on the local stack of 'T'

	TlContext
		ctxT1 = { TOP_OF_STACK(stackT1) },
		ctxT2 = { TOP_OF_STACK(stackT2) };

	tlCreate(&ctxT1,threadT1);
	tlCreate(&ctxT2,threadT2);


	printString("SP in T="); printHex32(getSp()); printLn();

	tlYield(ctx);
	while (true) {
		tlCall(&ctxT1);
		tlCall(&ctxT2);
		sleepMs(ctx,1000);
	}
}

void threadT1(TlContext *ctx) {
	while (true) {
		tlYield(ctx);
		printString("T1");
	}
}

void threadT2(TlContext *ctx) {
	while (true) {
		tlYield(ctx);
		printString("T2");
	}
}

void runThreads(void);

void main(void) {
	systickTimerInit(PCLK / 100/*Hz*/, 10/*ms*/ );

	SYSCON
	.ahbClkCtrl	|= 1<<_SYSCON_AHBCLKCTRL_SWM
			| 1<<_SYSCON_AHBCLKCTRL_USART0
			;

	swmPin(SWM_FUNCTION_U0_RXD,0);
	swmPin(SWM_FUNCTION_U0_TXD,4);

	usart0InitFrom12Mhz (1);
	print (usart0BlockingWriteChar);

	printString("Running.\n");
	printString("SP before="); printHex32(getSp()); printLn();

	runThreads();

	printString("\nSP after="); printHex32(getSp()); printLn();

	// all threads and their stacks gone by at this point...
	printString("Terminated.\n");
}

void runThreads(void) {
	Uint32	stackA[64];
	Uint32	stackB[64];
	Uint32	stackT[3*64];

	TlContext
		ctxA = { TOP_OF_STACK(stackA) },
		ctxB = { TOP_OF_STACK(stackB) },
		ctxT = { TOP_OF_STACK(stackT) };

	tlCreate(&ctxA,threadA);
	tlCreate(&ctxB,threadB);
	tlCreate(&ctxT,threadT);

	printString("SP in scheduler="); printHex32(getSp()); printLn();

	// Round robin scheduler, 10 seconds lifetime.
	for (int tMs=systickTimerMs+10*1000; tMs-systickTimerMs>=0; ) {
		tlCall(&ctxA);
		tlCall(&ctxB);
		tlCall(&ctxT);
	}
}

